package server

import (
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/getsentry/raven-go"
	gnatsd "github.com/nats-io/gnatsd/server"
	"github.com/nats-io/go-nats"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/auth"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// Options of a server
type Options struct {
	DatabaseDSN                    string
	DatabaseLogFlags               qb.LogFlags
	DatabaseMaxConnections         int
	ProcessStorageDBMaxConnections int
	RootCA                         string // a PEM or a path to a PEM file
	ServerCert                     string // a PEM or a path to a PEM file
	ServerKey                      string // a PEM or a path to a PEM file
	ClientCA                       string // a PEM or a path to a PEM file
	ClientCAKey                    string // a PEM or a path to a PEM file
	TLSTimeout                     float64
	NATSHost                       string
	NATSPort                       int
	NATSDebug                      bool
	NATSTrace                      bool
	ServeNATS                      bool
	ServeBroker                    bool
	ServeClientAPI                 bool
	ServeEmission                  bool
	ServeControl                   bool
	ServePostmortem                bool
	ServeProcessState              bool
	ServeRegistration              bool
	ServeSentinel                  bool
	GnatsdOptions                  gnatsd.Options
	SentryDSN                      string

	EmissionReceptionTimeout time.Duration
}

// Server the XBus server
type Server struct {
	options Options

	defaultModel        *model.Model
	defaultDB           *yago.DB
	processStorageModel *model.Model
	processStorageDB    *yago.DB

	accountStorage  storage.AccountStorage
	actorStorage    storage.ActorStorage
	graphStorage    storage.GraphStorage
	processStorage  storage.ProcessStorage
	envelopeStorage storage.EnvelopeStorage
	logStorage      storage.LogStorage

	processManager *process.Manager
	graphManager   *graph.Manager

	rootCA      *x509.Certificate
	serverCert  *x509.Certificate
	serverKey   *ecdsa.PrivateKey
	clientCA    *x509.Certificate
	clientCAKey *ecdsa.PrivateKey

	natsAuth   auth.XBusAuth
	natsServer *gnatsd.Server
	natsConn   *nats.EncodedConn

	ServiceManager *services.Manager

	ravenClient *raven.Client

	log        xbus.Logger
	natsLogger xbus.Logger
}

// NewServer creates a Server
func NewServer(options Options) *Server {
	if options.DatabaseMaxConnections == 0 {
		options.DatabaseMaxConnections = 10
	}
	if options.ProcessStorageDBMaxConnections == 0 {
		options.ProcessStorageDBMaxConnections = options.DatabaseMaxConnections/3 + 1
	}
	return &Server{
		options:    options,
		log:        xbus.NewNoopLogger(),
		natsLogger: xbus.NewNoopLogger(),
	}
}

func (s *Server) wrapLogger(logger xbus.Logger) xbus.Logger {
	if s.ravenClient != nil {
		s.ravenClient.Close()
		s.ravenClient = nil
	}
	if s.options.SentryDSN != "" {
		client, err := raven.New(s.options.SentryDSN)
		if err != nil {
			panic(err)
		}
		logger = xbus.NewRavenLogger(client, logger)
		client.Close()
	}
	return logger
}

// SetLogger changes the logger used by server
func (s *Server) SetLogger(logger xbus.Logger) {
	if logger == nil {
		logger = xbus.NewNoopLogger()
	}

	logger = s.wrapLogger(logger)

	s.log = xbus.NewPrefixedLogger("[server] ", logger)
	s.natsLogger = xbus.NewPrefixedLogger("[gnatsd] ", logger)
}

// SetSQLDB sets the model & db
func (s *Server) SetSQLDB(m *model.Model, db *yago.DB) {
	s.defaultModel = m
	s.defaultDB = db
}

// SetEnvelopeStorageDB sets the model & db for the envelope storage
func (s *Server) SetEnvelopeStorageDB(m *model.Model, db *yago.DB) {
	s.processStorageModel = m
	s.processStorageDB = db
}

func (s *Server) setupSQLDB() error {
	if s.defaultModel == nil {
		m, db, err := model.NewDB(
			"postgres", s.options.DatabaseDSN, s.options.DatabaseMaxConnections,
			xbus.NewPrefixedLogger("[sql] ", s.log),
			s.options.DatabaseLogFlags)
		if err != nil {
			s.log.Errorf("Cannot connect to the database.Got err %s", err)
			return err
		}
		s.SetSQLDB(m, db)
	}
	if s.processStorageModel == nil {
		m, db, err := model.NewDB(
			"postgres", s.options.DatabaseDSN, s.options.ProcessStorageDBMaxConnections,
			xbus.NewPrefixedLogger("[sql(processstorage)] ", s.log),
			s.options.DatabaseLogFlags)
		if err != nil {
			s.log.Errorf("Cannot connect to the database.Got err %s", err)
			return err
		}
		s.SetEnvelopeStorageDB(m, db)
	}
	return nil
}

// AccountStorage returns the underlying AccountStorage
func (s *Server) AccountStorage() storage.AccountStorage {
	return s.accountStorage
}

// ActorStorage returns the underlying ActorStorage
func (s *Server) ActorStorage() storage.ActorStorage {
	return s.actorStorage
}

// ProcessStorage returns the underlying ProcessStorage
func (s *Server) ProcessStorage() storage.ProcessStorage {
	return s.processStorage
}

// ProcessManager returns the process manager
func (s *Server) ProcessManager() *process.Manager {
	return s.processManager
}

func (s *Server) setupStorage() {
	s.accountStorage = sql.NewAccountStorage(s.defaultModel, s.defaultDB, s.log)
	s.actorStorage = sql.NewActorStorage(s.defaultModel, s.defaultDB, s.log)
	s.envelopeStorage = sql.NewEnvelopeStorage(s.defaultModel, s.defaultDB, s.log)
	s.graphStorage = sql.NewGraphStorage(s.defaultModel, s.defaultDB, s.log)
	s.processStorage = sql.NewProcessStorage(s.processStorageModel, s.processStorageDB, s.log)
	s.logStorage = sql.NewLogStorage(s.defaultModel, s.defaultDB, s.log)
}

func (s *Server) setupManagers() {
	s.graphManager = graph.NewManager(
		s.actorStorage, s.graphStorage)
	s.processManager = process.NewManager(
		s.graphManager, s.actorStorage, s.processStorage, s.log,
	)
}

func (s *Server) setupCrypto() error {
	rootCA, err := tools.ReadPEMCertificate(s.options.RootCA)
	if err != nil {
		return err
	}
	serverCert, err := tools.ReadPEMCertificate(s.options.ServerCert)
	if err != nil {
		return err
	}
	serverKey, err := tools.ReadPEMPrivateKey(s.options.ServerKey)
	if err != nil {
		return err
	}
	clientCA, err := tools.ReadPEMCertificate(s.options.ClientCA)
	if err != nil {
		return err
	}
	clientCAKey, err := tools.ReadPEMPrivateKey(s.options.ClientCAKey)
	if err != nil {
		return err
	}
	s.rootCA = rootCA
	s.serverCert = serverCert
	s.serverKey = serverKey
	s.clientCA = clientCA
	s.clientCAKey = clientCAKey

	return nil
}

// Startup starts the server
func (s *Server) Startup() (err error) {
	defer func() {
		if err != nil {
			s.Shutdown()
		}
	}()
	if err = s.setupSQLDB(); err != nil {
		return err
	}
	s.setupStorage()
	s.setupManagers()
	if err = s.setupCrypto(); err != nil {
		return err
	}
	if s.options.ServeNATS {
		if err = s.setupNATS(); err != nil {
			return err
		}
		if err = s.StartNATS(); err != nil {
			return err
		}
	}
	s.setupServices()
	s.log.Noticef("Starting services")
	if err = s.ServiceManager.Startup(); err != nil {
		return err
	}
	return nil
}

// Shutdown stops the services and nats server (if relevant)
func (s *Server) Shutdown() {
	if s.ServiceManager != nil {
		if err := s.ServiceManager.Shutdown(); err != nil {
			s.log.Errorf("Error shutting down services: %s", err)
		}
	}
	if s.options.ServeNATS {
		s.StopNATS()
	}
}

func (s *Server) setupNATS() error {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.clientCA)

	s.natsAuth = auth.NewXBusAuth(s.accountStorage, s.actorStorage, s.log)

	gnatsdOptions := s.options.GnatsdOptions
	gnatsdOptions.Host = s.options.NATSHost
	gnatsdOptions.Port = s.options.NATSPort
	gnatsdOptions.NoSigs = true
	gnatsdOptions.TLSConfig = &tls.Config{
		Certificates: []tls.Certificate{
			tls.Certificate{
				Certificate: [][]byte{s.serverCert.Raw},
				PrivateKey:  s.serverKey,
				Leaf:        s.serverCert,
			},
		},
		ClientAuth: tls.VerifyClientCertIfGiven,
		ClientCAs:  certPool,
	}
	gnatsdOptions.CustomClientAuth = &s.natsAuth
	gnatsdOptions.TLSTimeout = s.options.TLSTimeout

	s.natsServer = gnatsd.New(&gnatsdOptions)
	s.natsServer.SetLogger(s.natsLogger, s.options.NATSDebug, s.options.NATSTrace)

	return nil
}

// getNATSPort returns the nats server port, when an
// error occurs the returned value should be ignored.
func (s *Server) getNATSPort() (int, error) {
	endpoint := s.natsServer.Addr()
	if endpoint == nil {
		return 0, errors.New("Undefined endpoint")
	}
	subStr := strings.SplitAfter(endpoint.String(), ":")
	strPort := subStr[len(subStr)-1]
	port, err := strconv.Atoi(strPort)
	if err != nil {
		return 0, err
	}
	return port, nil
}

// NATSUrl returns the nats server url
func (s *Server) NATSUrl() string {
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	return fmt.Sprintf("nats://%s:%d", s.options.NATSHost, port)
}

// NATSConnectOptions returns nats.Options to connect to the
// bus with full access
func (s *Server) NATSConnectOptions() nats.Options {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.rootCA)

	return nats.Options{
		Url:   s.NATSUrl(),
		Token: s.natsAuth.AuthToken,
		TLSConfig: &tls.Config{
			RootCAs: certPool,
		},
		Secure: true,
	}
}

func (s *Server) waitForNATS() error {

	if !s.natsServer.ReadyForConnections(time.Second) {
		return errors.New("Server still has no listen endpoint")
	}
	url := s.NATSUrl()
	if url == "" {
		return errors.New("Server still has no listen endpoint")
	}

	/*
		opts := s.NATSConnectOptions()

		var (
			ready   bool
			lastErr error
		)
		for time.Now().Before(timeout) {
			// let the goroutine start to listen
			runtime.Gosched()

			natsConn, err := opts.Connect()
			lastErr = err
			if err == nil {
				natsConn.Close()
				ready = true
				break
			}
		}
		if !ready {
			return fmt.Errorf("Cannot connect to the nats server: %s", lastErr)
		}
	*/
	s.log.Noticef("Nats server responding on %s", s.NATSUrl())

	// Store the actual port back into the options so that a restart
	// will reuse the same port, even if the original value is -1 which means
	// a random port.
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	s.options.NATSPort = port
	return nil
}

// StartNATS starts the nats server
func (s *Server) StartNATS() error {
	s.log.Noticef("Starting NATS server")

	go s.natsServer.Start()

	if err := s.waitForNATS(); err != nil {
		s.StopNATS()
		return err
	}

	connOptions := nats.Options{
		Url:    s.NATSUrl(),
		Token:  s.natsAuth.AuthToken,
		Secure: true,
		AsyncErrorCB: func(c *nats.Conn, sub *nats.Subscription, err error) {
			if sub != nil {
				s.log.Errorf("Error on sub %s: %s", sub.Subject, err)
			} else {
				s.log.Errorf("Error %s", err)
			}
		},
	}

	conn, err := connOptions.Connect()

	if err != nil {
		s.log.Errorf("Could not connect to the NATS server: %s", err)
		s.StopNATS()
		return err
	}

	eConn, err := nats.NewEncodedConn(conn, "json")

	if err != nil {
		conn.Close()
		s.StopNATS()
		return fmt.Errorf("Could not initialize the encoded connection: %s", err)
	}

	s.natsConn = eConn

	return nil
}

// StopNATS stops the nats server
func (s *Server) StopNATS() {
	if s.natsConn != nil {
		s.natsConn.Close()
		s.natsConn = nil
	}
	if s.natsServer != nil {
		s.natsServer.Shutdown()
		s.natsServer = nil
	}
}
