package core

import (
	"testing"

	"github.com/m4rw3r/uuid"
	_ "github.com/slicebit/qb/dialects/postgres"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql"
	"bitbucket.org/orus-io/xbus/testutils"
	"bitbucket.org/orus-io/xbus/testutils/sqldb"
)

func TestUpdateState(t *testing.T) {
	logger := testutils.GetTestLogger(t)

	testdb := sqldb.GetTestDB(t)
	defer testdb.Close()
	model, db := testdb.GetModelDB(1)

	actorStorage := sql.NewActorStorage(model, db, logger)
	graphStorage := sql.NewGraphStorage(model, db, logger)
	graphManager := graph.NewManager(actorStorage, graphStorage)
	processStorage := sql.NewProcessStorage(model, db, logger)
	envelopeStorage := sql.NewEnvelopeStorage(model, db, logger)

	graphID, err := graphStorage.Save(storage.Graph{
		Name:    "default",
		Version: "0.1",
		Nodes: []storage.Node{
			storage.Node{
				ID:            "source",
				Type:          storage.EmitterNode,
				Roles:         []string{"default"},
				RoleBroadcast: true,
			},
			storage.Node{
				ID:            "sink",
				Type:          storage.ConsumerNode,
				Roles:         []string{"default"},
				RoleBroadcast: true,
			},
		},
		Edges: []storage.Edge{
			storage.Edge{
				From: storage.NodeOutputRef{NodeID: "source", Output: "default"},
				To:   storage.NodeInputRef{NodeID: "sink", Input: "default"},
			},
		},
	})
	if err != nil {
		t.Fatal(err)
	}
	if err := graphStorage.SetStatus(graphID, api.GraphActive); err != nil {
		t.Fatal(err)
	}

	emitter, err := actorStorage.Create(api.Actor{
		Name:  "emitter",
		Kind:  api.EmitterKind,
		Roles: []string{"default"},
	})
	if err != nil {
		t.Fatal(err)
	}
	emID := emitter.ID

	envID := uuid.MustFromString("2dd179e8-b40b-11e7-9aad-7fcecdf65470")
	evID := uuid.MustFromString("df854950-b40e-11e7-a515-bb760274e30c")

	if _, err := envelopeStorage.StoreEnvelope(api.Envelope{
		ID:       envID,
		EventIDs: []uuid.UUID{evID},
		Events: []api.Event{
			{
				ID:        evID,
				Items:     []api.Item{},
				ItemCount: 0,
				Checksum:  0,
			},
		},
		Last: true,
	}); err != nil {
		t.Fatal(err)
	}

	service := Options{
		ActorStorage:   actorStorage,
		ProcessStorage: processStorage,
		ProcessManager: process.NewManager(graphManager, actorStorage, processStorage, logger),
		GraphManager:   graphManager,
		LogStorage:     sql.NewLogStorage(model, db, logger),
		Options: services.Options{
			Logger: logger,
		},
	}.NewService()

	var logentries []storage.ProcessLogEntry

	processStorage.SetProcessLogEntryHandler(func(entries []storage.ProcessLogEntry) {
		logentries = append(logentries, entries...)
	})

	pID, err := processStorage.CreateProcess(storage.Process{
		TriggerEmitterID:  emID,
		TriggerEnvelopeID: envID,
		GraphID:           graphID,
		SourceOutput:      storage.NodeOutputRef{NodeID: "source", Output: "default"},
	})
	if err != nil {
		t.Fatal(err)
	}

	if err := processStorage.UpdateReceiveStatus(
		storage.Reception{
			ProcessID:   pID,
			EnvelopeID:  envID,
			SourceActor: emID,
			SourceOutputRef: storage.NodeOutputRef{
				NodeID: "source",
				Output: "default",
			},
		},
		storage.ReceptionDone,
	); err != nil {
		t.Fatal(err)
	}

	if err := processStorage.ProcessNodeEnd(
		pID, storage.ProcessNode{ActorID: emID, NodeID: "sink"}, emID, api.ActorProcessingSuccess,
	); err != nil {
		t.Fatal(err)
	}

	logger.ExpectsError(testutils.ExpectsContains("Error calculating next state: "))
	for _, entry := range logentries {
		service.updateState(entry.ProcessID, []storage.ProcessLogEntry{entry})
	}

	state := processStorage.GetState(pID)
	assert.Equal(t, api.ProcessError, state.Status)
	assert.Contains(t, state.StatusReason, "Could not compute next state: Error calculating next state:")

	var count int
	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(false),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 0, count)

	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(true),
		model.ProcessLogEntry.ProcessingError.Eq(false),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 1, count)

	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(true),
		model.ProcessLogEntry.ProcessingError.Eq(true),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 1, count)
}
