package storage

import (
	"errors"

	"bitbucket.org/orus-io/go-xbus/api"

	"github.com/m4rw3r/uuid"
)

var (
	// ErrInvalidGraphStatus is returned when the graph status does
	// allow the operation
	ErrInvalidGraphStatus = errors.New("Invalid Graph Status")
)

// NodeType is the type of Node
//go:generate enum-marshaler -type NodeType -suffix Node -lower graph.go
type NodeType int

const (
	// EmitterNode is the type of emitter nodes
	EmitterNode NodeType = iota
	// ConsumerNode is the type of consumer nodes
	ConsumerNode
	// WorkerNode is the type of worker nodes
	WorkerNode
)

// SourceMatch are rules to match incoming envelope with
type SourceMatch struct {
	EventTypes []string // All event types envelope must contain exactly
}

// Node is a graph vertex
type Node struct {
	ID            string
	Type          NodeType
	ActorIDs      []uuid.UUID // Explicit actors. Deprecated (will be removed in a later version)
	Actors        []string    // Explicit actors, by name or UUID
	Roles         []string    // Actor roles
	RoleBroadcast bool        // If true, all actors matching each role will receive the message, instead of only one actor per role

	SourceMatch *SourceMatch // define only if Type==EmitterNode

	Inputs  []string
	Outputs []string

	// TODO add 'arguments'
}

// NodeInputRef is a reference to a node input
type NodeInputRef struct {
	NodeID string
	Input  string
}

// NodeOutputRef is a reference to a node input
type NodeOutputRef struct {
	NodeID string
	Output string
}

// Edge is a Graph oriented edge
type Edge struct {
	From NodeOutputRef
	To   NodeInputRef
}

// Graph describe a processing graph.
type Graph struct {
	ID      uuid.UUID
	Name    string
	Version string
	Status  api.GraphStatus
	Nodes   []Node
	Edges   []Edge
}

// GraphStorage handle Graph persistence
type GraphStorage interface {
	Load(uuid.UUID) (Graph, bool)
	Save(Graph) (uuid.UUID, error)
	SetStatus(uuid.UUID, api.GraphStatus) error
	LoadActiveGraph() []Graph
	Query(name, version string, activeOnly bool) []Graph
}
