package sql

import (
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

func readStringList(s string) []string {
	if s == "" {
		return []string{}
	}
	return strings.Split(s, ",")
}

func writeStringList(list []string) string {
	return strings.Join(list, ",")
}

// NewGraphStorage initialize a GraphStorage
func NewGraphStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *GraphStorage {
	return &GraphStorage{Base{
		model:  model,
		db:     db,
		logger: logger,
	}}
}

// GraphStorage implements storage.GraphStorage
type GraphStorage struct {
	Base
}

func dbNodeToStorage(dbNode model.GraphNode) storage.Node {
	node := storage.Node{
		ID:            dbNode.NodeID,
		RoleBroadcast: dbNode.RoleBroadcast,
		Inputs:        readStringList(dbNode.Inputs),
		Outputs:       readStringList(dbNode.Outputs),
	}
	switch dbNode.Type {
	case "emitter":
		node.Type = storage.EmitterNode
		node.SourceMatch = &storage.SourceMatch{
			EventTypes: strings.Split(dbNode.SourceMatchEventTypes, ","),
		}
	case "consumer":
		node.Type = storage.ConsumerNode
	case "worker":
		node.Type = storage.WorkerNode
	}
	for _, id := range strings.Split(dbNode.ActorIDs, ",") {
		if id != "" {
			node.ActorIDs = append(node.ActorIDs, uuid.MustFromString(id))
		}
	}
	for _, actor := range strings.Split(dbNode.Actors, ",") {
		if actor != "" {
			node.Actors = append(node.Actors, actor)
		}
	}
	if dbNode.Roles != "" {
		node.Roles = strings.Split(dbNode.Roles, ",")
	}
	return node
}

func parseEdgeList(str string) ([]storage.Edge, error) {
	var edges []storage.Edge
	for _, strEdge := range strings.Split(str, ",") {
		if strEdge == "" {
			continue
		}
		edge, err := storage.EdgeFromString(strEdge)
		if err != nil {
			return nil, err
		}
		edges = append(edges, edge)
	}
	return edges, nil
}

func writeEdgeList(edges []storage.Edge) string {
	var strEdges []string
	for _, edge := range edges {
		strEdges = append(strEdges, edge.String())
	}
	return strings.Join(strEdges, ",")
}

func dbGraphToStorage(dbGraph model.Graph, dbNodes []model.GraphNode) (storage.Graph, error) {
	var nodes []storage.Node
	for _, dbNode := range dbNodes {
		nodes = append(nodes, dbNodeToStorage(dbNode))
	}
	edges, err := parseEdgeList(dbGraph.Edges)
	if err != nil {
		return storage.Graph{}, err
	}
	graph := storage.Graph{
		ID:      dbGraph.ID,
		Name:    dbGraph.Name,
		Version: dbGraph.Version,
		Status:  dbGraph.Status,
		Nodes:   nodes,
		Edges:   edges,
	}
	return graph, nil
}

var storageNodeTypeToDB = []string{"emitter", "consumer", "worker"}

func storageNodeToDB(node storage.Node) model.GraphNode {
	var strActorIDs []string
	for _, id := range node.ActorIDs {
		strActorIDs = append(strActorIDs, id.String())
	}

	var sourceMatchEventTypes string
	if node.Type == storage.EmitterNode && node.SourceMatch != nil {
		sourceMatchEventTypes = strings.Join(node.SourceMatch.EventTypes, ",")
	}

	return model.GraphNode{
		NodeID:                node.ID,
		Type:                  storageNodeTypeToDB[node.Type],
		ActorIDs:              strings.Join(strActorIDs, ","),
		Actors:                strings.Join(node.Actors, ","),
		Roles:                 writeStringList(node.Roles),
		RoleBroadcast:         node.RoleBroadcast,
		SourceMatchEventTypes: sourceMatchEventTypes,
		Inputs:                writeStringList(node.Inputs),
		Outputs:               writeStringList(node.Outputs),
	}
}

func storageGraphToDB(graph storage.Graph) (model.Graph, []model.GraphNode) {
	var dbNodes []model.GraphNode
	for _, node := range graph.Nodes {
		dbNodes = append(dbNodes, storageNodeToDB(node))
	}
	return model.Graph{
		Base:    model.Base{ID: graph.ID},
		Name:    graph.Name,
		Version: graph.Version,
		Status:  graph.Status,
		Edges:   writeEdgeList(graph.Edges),
	}, dbNodes
}

// Load loads a graph from storage
func (s *GraphStorage) Load(id uuid.UUID) (storage.Graph, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbGraph model.Graph
	err := tx.Query(s.model.Graph).Get(&dbGraph, id)
	if err == yago.ErrRecordNotFound {
		return storage.Graph{}, false
	}
	must(err)

	var dbNodes []model.GraphNode
	must(
		tx.Query(s.model.GraphNode).
			Where(s.model.GraphNode.GraphID.Eq(id)).
			All(&dbNodes))

	graph, err := dbGraphToStorage(dbGraph, dbNodes)
	must(err)

	return graph, true
}

// Save saves a new graph to storage
func (s *GraphStorage) Save(graph storage.Graph) (uuid.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbGraph model.Graph

	if !graph.ID.IsZero() {
		if err := tx.Query(s.model.Graph).
			Where(s.model.Graph.ID.Eq(graph.ID)).
			One(&dbGraph); err != nil {
			if err == yago.ErrRecordNotFound {
				err = fmt.Errorf("No such graph: %s", graph.ID)
				return uuid.UUID{}, err
			}
			must(err)
			if dbGraph.Status != api.GraphDraft {
				return uuid.UUID{}, storage.ErrInvalidGraphStatus
			}
		}
	} else {
		if err := tx.Query(s.model.Graph).
			Where(
				s.model.Graph.Name.Eq(graph.Name),
				s.model.Graph.Version.Eq(graph.Version)).
			One(&dbGraph); err != nil {
			if err != yago.ErrRecordNotFound {
				must(err)
			}
			if err == nil && dbGraph.Status != api.GraphDraft {
				return uuid.UUID{}, storage.ErrInvalidGraphStatus
			}
		}
	}

	isNewGraph := dbGraph.ID.IsZero()

	nextGraph, dbNodes := storageGraphToDB(graph)

	dbGraph.Name = nextGraph.Name
	dbGraph.Version = nextGraph.Version
	dbGraph.Edges = nextGraph.Edges

	if isNewGraph {
		s.logger.Noticef("Inserting graph")
		must(
			tx.Insert(&dbGraph))
	} else {
		s.logger.Noticef("Updating graph")
		must(
			tx.Update(
				&dbGraph,
				model.GraphName,
				model.GraphVersion,
			))
	}

	s.logger.Noticef("Got %s", dbGraph.ID, dbNodes)
	var dbNodeIDs []interface{}

	q := tx.Query(s.model.GraphNode).
		Select(s.model.GraphNode.ID).
		Where(s.model.GraphNode.GraphID.Eq(dbGraph.ID))

	for _, dbNode := range dbNodes {
		dbNode.GraphID = dbGraph.ID
		var id int64
		if err := q.Filter(
			s.model.GraphNode.NodeID.Eq(dbNode.NodeID),
		).Scalar(&id); err != nil {
			if err == yago.ErrRecordNotFound {
				s.logger.Noticef("Insert node %d (%s)", id, dbNode)
				must(
					tx.Insert(&dbNode))
			} else {
				must(err)
			}
		} else {
			dbNode.ID = id
			s.logger.Noticef("Update node %d (%s)", id, dbNode)
			must(
				tx.Update(&dbNode))
		}
		dbNodeIDs = append(dbNodeIDs, dbNode.NodeID)
	}

	_, err := tx.GetEngine().Exec(
		qb.Delete(*s.model.GraphNode.GetMapper().Table()).
			Where(qb.And(
				s.model.GraphNode.GraphID.Eq(dbGraph.ID),
				s.model.GraphNode.NodeID.NotIn(dbNodeIDs...))),
	)
	must(err)
	tx.MustCommit()

	return dbGraph.ID, nil
}

// SetStatus changes a graph status
func (s *GraphStorage) SetStatus(id uuid.UUID, status api.GraphStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	if status != api.GraphActive && status != api.GraphInactive {
		return storage.ErrInvalidGraphStatus
	}

	if status == api.GraphActive {
		// Deactivate any other version of the same graph
		var name string
		must(tx.Query(s.model.Graph).
			Where(s.model.Graph.ID.Eq(id)).
			Select(s.model.Graph.Name).
			Scalar(&name))

		_, err := tx.GetEngine().Exec(
			s.model.Graph.GetMapper().Table().Update().
				Values(map[string]interface{}{
					model.GraphStatusColumnName: "inactive",
				}).
				Where(qb.And(
					s.model.Graph.Name.Eq(name),
					s.model.Graph.Status.Eq("active"),
				)),
		)
		must(err)
	}

	must(tx.Update(
		&model.Graph{Base: model.Base{ID: id}, Status: status},
		model.GraphStatus))

	tx.MustCommit()

	return nil
}

// LoadActiveGraph loads all the active graphs
func (s *GraphStorage) LoadActiveGraph() []storage.Graph {
	tx := s.MustBegin()
	defer tx.Rollback()

	var dbGraphs []model.Graph
	must(
		tx.Query(s.model.Graph).
			Where(s.model.Graph.Status.Eq("active")).
			All(&dbGraphs))

	var graphList []storage.Graph

	for _, dbGraph := range dbGraphs {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.GraphID.Eq(dbGraph.ID)).
				All(&dbNodes))
		graph, err := dbGraphToStorage(dbGraph, dbNodes)
		must(err)
		graphList = append(graphList, graph)
	}
	return graphList
}

// Query searches for matching graphs
func (s *GraphStorage) Query(name, version string, activeOnly bool) (graphs []storage.Graph) {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Graph)
	if name != "" {
		q = q.Filter(s.model.Graph.Name.Eq(name))
	}
	if version != "" {
		q = q.Filter(s.model.Graph.Version.Eq(version))
	}
	if activeOnly {
		q = q.Filter(s.model.Graph.Status.Eq(api.GraphActive))
	}

	var dbGraphs []model.Graph
	must(q.All(&dbGraphs))

	for _, dbGraph := range dbGraphs {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.GraphID.Eq(dbGraph.ID)).
				All(&dbNodes))

		graph, err := dbGraphToStorage(dbGraph, dbNodes)
		must(err)

		graphs = append(graphs, graph)
	}
	return
}
