// Copyright 2017 Orus.io. All rights reserved.

package sql

import (
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	xuuid "bitbucket.org/orus-io/xbus/uuid"
)

// NewLogStorage initialize a LogStorage
func NewLogStorage(model *model.Model, db *yago.DB, logger xbus.Logger) LogStorage {
	return LogStorage{Base{
		model:  model,
		db:     db,
		logger: xbus.NewPrefixedLogger("[storage/sql/envelop]", logger),
	}}
}

// LogStorage implements storage.LogStorage
type LogStorage struct {
	Base
}

// Log adds a log entry to the storage
func (s LogStorage) Log(entry api.LogEntry) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntry = model.LogEntry{
		EnvelopeID: xuuid.NullUUID(entry.EnvelopeID),
		ActorID:    xuuid.NullUUID(entry.ActorID),
		ProcessID:  xuuid.NullUUID(entry.ProcessID),
		NodeID:     entry.NodeID,
		Time:       entry.Message.Time,
		Level:      entry.Message.Level,
		Text:       entry.Message.Text,
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		panic(err)
	}
	tx.MustCommit()
}

func (s LogStorage) list(crits ...qb.Clause) []api.LogEntry {
	tx := s.MustBegin()
	defer tx.Rollback()

	var dbEntries []model.LogEntry
	if err := tx.Query(s.model.LogEntry).Where(crits...).All(&dbEntries); err != nil {
		panic(err)
	}

	var entries []api.LogEntry
	for _, dbEntry := range dbEntries {
		entries = append(entries, api.LogEntry{
			EnvelopeID: uuid.UUID(dbEntry.EnvelopeID),
			ActorID:    uuid.UUID(dbEntry.ActorID),
			ProcessID:  uuid.UUID(dbEntry.ProcessID),
			NodeID:     dbEntry.NodeID,
			Message: api.LogMessage{
				Time:  dbEntry.Time,
				Level: dbEntry.Level,
				Text:  dbEntry.Text,
			},
		})
	}
	return entries
}

// GetProcessMessages returns all the messages related to the given envelope
func (s LogStorage) GetProcessMessages(processID uuid.UUID, level api.LogLevel) []api.LogEntry {
	var crits = []qb.Clause{
		s.model.LogEntry.ProcessID.Eq(processID),
	}

	switch level {
	case api.LogError:
		crits = append(crits, s.model.LogEntry.Level.Eq(level))
	case api.LogWarning:
		crits = append(crits, s.model.LogEntry.Level.NotEq(api.LogNotice))
	}

	return s.list(crits...)
}

// PurgeBefore removes all logs older than the given date
func (s LogStorage) PurgeBefore(t time.Time) int64 {
	tx := s.MustBegin()
	defer tx.MustCommit()

	res, err := tx.GetEngine().Exec(
		s.model.LogEntry.GetMapper().Table().Delete().Where(
			s.model.LogEntry.Time.Lt(t)))
	must(err)
	ra, err := res.RowsAffected()
	must(err)
	return ra
}
