// Copyright 2016 Orus.io. All rights reserved.

package sqldb

import (
	"fmt"
	"os"
	"testing"

	// load postgres driver
	_ "github.com/lib/pq"
	// load sqlite driver
	_ "github.com/mattn/go-sqlite3"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"
	"github.com/tgulacsi/go-locking"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

const testMaxDBConnections = 10

// CloseFunction is a function with no arg nor return values for closing a resource
type CloseFunction func()

// TestDB holds a test database and provides access to it until its 'Close'
// function is called
type TestDB struct {
	Driver   string
	Dsn      string
	LogFlags qb.LogFlags
	Logger   xbus.Logger

	t    testing.TB
	lock *locking.PortLock
}

// Close releases the underlying database
func (tdb *TestDB) Close() {
	if tdb.lock != nil {
		if err := tdb.lock.Unlock(); err != nil {
			tdb.t.Errorf("Error releasing lock: %s", err)
		}
	}
}

// GetModelDB returns a new connection pool to the database
func (tdb *TestDB) GetModelDB(maxConn int) (*model.Model, *yago.DB) {
	m, db, err := model.NewDB(tdb.Driver, tdb.Dsn, maxConn, tdb.Logger, tdb.LogFlags)
	if err != nil {
		tdb.t.Fatalf("Cannot initialize test database: %s", err)
		return nil, nil
	}
	return m, db
}

// GetTestDB creates a test database and returns a TestDB pointing to it
func GetTestDB(t testing.TB) *TestDB {
	var (
		driver     = "sqlite3"
		droptables = false
	)

	dsn, ok := os.LookupEnv("XBUS_TEST_POSTGRES_DSN")
	if ok {
		driver = "postgres"
		droptables = true
	} else {
		dsn, ok = os.LookupEnv("XBUS_TEST_SQLITE_DSN")
		if !ok {
			os.Remove("test.db")
			dsn = "test.db"
		}
	}
	logFlags := qb.LDefault
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGQUERY"); v != "" {
		logFlags |= qb.LQuery
	}
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGBINDINGS"); v != "" {
		logFlags |= qb.LBindings
	}
	tdb := TestDB{
		Driver: driver,
		Dsn:    dsn,
		Logger: testutils.GetTestLogger(t),
		t:      t,
	}
	_, db := tdb.GetModelDB(1)
	defer db.Close()

	var allGood bool

	defer func() {
		if !allGood {
			tdb.Close()
		}
	}()
	if driver == "postgres" {
		lock := locking.NewPortLock(1764)
		err := lock.Lock()
		if err != nil {
			t.Errorf("Cannot get lock: %s", err)
			return nil
		}
		tdb.lock = lock
	}
	if droptables {
		tables := db.Metadata.GetQbMetadata().Tables()
		for i := len(tables) - 1; i >= 0; i-- {
			table := tables[i]
			drop := table.Drop(db.Engine.Dialect())
			if _, err := db.Engine.DB().Exec(drop); err != nil {
				t.Log(err)
			}
		}
	}

	if err := db.Metadata.GetQbMetadata().CreateAll(db.Engine); err != nil {
		fmt.Println(err)
		t.Fatal(err)
	}
	allGood = true
	return &tdb
}
