// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"errors"
	"fmt"
	"strings"
	"time"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// ErrNotFound Not match found
var ErrNotFound = errors.New("Not found")

// ValidationMessagesList A list of validation messages that implements
// the 'error' interface
type ValidationMessagesList []string

// Error return the combined list of messages in one string
func (l ValidationMessagesList) Error() string {
	return "Graph Validation error: \n* " + strings.Join(l, "\n* ")
}

func wrapAPIError(res api.Response) error {
	return fmt.Errorf("%s: %s", res.Status, errors.New(res.StatusMessage))
}

func (c *Client) actorOp(action string, actors []api.Actor) ([]api.Actor, error) {
	var res api.ControlActorResponse
	err := c.conn.Request(api.ControlActorsPrefix+action, api.ControlActorRequest{Actors: actors}, &res, 5*time.Second)
	if err != nil {
		return nil, err
	}
	if res.Status != api.StatusOK {
		return nil, wrapAPIError(res.Response)
	}
	return res.Actors, nil
}

// ActorCreate create actors
func (c *Client) ActorCreate(actors []api.Actor) ([]api.Actor, error) {
	return c.actorOp(api.ActionCreate, actors)
}

// ActorDelete deletes actors
func (c *Client) ActorDelete(ids []uuid.UUID) error {
	var actors []api.Actor
	for _, id := range ids {
		actors = append(actors, api.Actor{ID: id})
	}
	_, err := c.actorOp(api.ActionDelete, actors)
	return err
}

// ActorList lists actors.
func (c *Client) ActorList(actors []api.Actor) ([]api.Actor, error) {
	return c.actorOp(api.ActionQuery, actors)
}

// ActorFind finds an actor
func (c *Client) ActorFind(identifier string) (api.Actor, error) {
	var filter api.Actor
	if id, err := uuid.FromString(identifier); err == nil {
		filter.ID = id
	} else {
		filter.Name = identifier
	}
	list, err := c.ActorList([]api.Actor{filter})
	if err != nil {
		return api.Actor{}, err
	}

	if len(list) == 0 {
		return api.Actor{}, ErrNotFound
	}
	if len(list) > 1 {
		panic("received multiple actors for " + identifier)
	}
	return list[0], nil
}

// ActorAccept accepts a pending actor
func (c *Client) ActorAccept(id uuid.UUID) (api.Actor, error) {
	var rep api.ControlActorResponse
	if err := c.conn.Request(
		api.ControlActorsUpdate,
		api.ControlActorRequest{
			Actors: []api.Actor{api.Actor{
				ID:     id,
				Status: api.ActorEnabled,
			}},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Actor{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Actor{}, wrapAPIError(rep.Response)
	}
	return rep.Actors[0], nil
}

// ActorReject rejects a pending actor
func (c *Client) ActorReject(id uuid.UUID) (api.Actor, error) {
	var rep api.ControlActorResponse
	if err := c.conn.Request(
		api.ControlActorsUpdate,
		api.ControlActorRequest{
			Actors: []api.Actor{api.Actor{
				ID:     id,
				Status: api.ActorRejected,
			}},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Actor{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Actor{}, wrapAPIError(rep.Response)
	}
	return rep.Actors[0], nil
}

// EnvelopePurge removes unreferenced envelopes
func (c *Client) EnvelopePurge() (int, error) {
	var rep api.ControlEnvelopePurgeRep
	if err := c.conn.Request(
		api.ControlEnvelopePurge,
		api.ControlEnvelopePurgeReq{},
		&rep, 5*time.Second,
	); err != nil {
		return 0, err
	}
	if rep.Status != api.StatusOK {
		return 0, wrapAPIError(rep.Response)
	}
	return rep.Count, nil
}

// GraphQuery searches for graphs
func (c *Client) GraphQuery(name string) ([]api.GraphInfo, error) {
	var rep api.ControlGraphQueryRep
	if err := c.conn.Request(api.ControlGraphQuery, api.ControlGraphQueryReq{
		Name: name,
	}, &rep, 5*time.Second); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	var graphs []api.GraphInfo
	for _, graph := range rep.GraphList {
		graphs = append(graphs, graph)
	}
	return graphs, nil
}

// GraphSave saves a draft graph to the server
func (c *Client) GraphSave(
	info api.GraphInfo, graph string,
) (api.GraphInfo, string, []string, error) {
	var rep api.ControlGraphSaveRep
	if err := c.conn.Request(
		api.ControlGraphSave,
		api.ControlGraphSaveReq{Info: info, Graph: graph},
		&rep, 5*time.Second,
	); err != nil {
		return info, "", nil, err
	}
	if rep.Status != api.StatusOK {
		return info, "", rep.ValidationMessages, wrapAPIError(rep.Response)
	}
	return rep.Info, rep.Graph, rep.ValidationMessages, nil
}

// GraphSetStatus changes the status of a graph
func (c *Client) GraphSetStatus(info api.GraphInfo) (api.GraphInfo, error) {
	var rep api.ControlGraphSetStatusRep
	if err := c.conn.Request(
		api.ControlGraphSetStatus, info,
		&rep, 5*time.Second,
	); err != nil {
		return info, err
	}
	if rep.Status != api.StatusOK {
		if rep.Status == api.StatusClientError && rep.StatusMessage == "invalidgraph" {
			return info, ValidationMessagesList(rep.ValidationMessages)
		}
		return info, wrapAPIError(rep.Response)
	}
	return rep.Info, nil
}

// ProcessQuery queries processes
func (c *Client) ProcessQuery(filter api.ProcessFilter) ([]api.Process, error) {
	var rep api.ControlProcessQueryRep
	if err := c.conn.Request(
		api.ControlProcessQuery,
		api.ControlProcessQueryReq{Filter: filter},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.ProcessList, nil
}

// ProcessControl pause/resume/cancel a process
func (c *Client) ProcessControl(processID uuid.UUID, command string) error {
	var rep api.ControlProcessControlRep
	if err := c.conn.Request(
		api.ControlProcessControl,
		api.ControlProcessControlReq{
			ProcessID: processID,
			Command:   command,
		},
		&rep, 5*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

// ProcessExport exports processes data (definition, state, logs, envelopes...)
func (c *Client) ProcessExport(processIDs []uuid.UUID) ([]string, error) {
	var rep api.ControlProcessExportRep
	if err := c.conn.Request(
		api.ControlProcessExport,
		api.ControlProcessExportReq{
			ProcessIDs: processIDs,
		},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.ProcessList, nil
}

// ProcessPurge purges processes data (definition, state, routes, logs)
func (c *Client) ProcessPurge(processIDs []uuid.UUID) error {
	var rep api.ControlProcessPurgeRep
	if err := c.conn.Request(
		api.ControlProcessPurge,
		api.ControlProcessPurgeReq{
			ProcessIDs: processIDs,
		},
		&rep, 5*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(api.Response(rep))
	}
	return nil
}

// AccountList lists the bus accounts
func (c *Client) AccountList() ([]api.Account, error) {
	var rep api.ControlAccountListRep
	if err := c.conn.Request(
		api.ControlAccountList,
		api.ControlAccountListReq{},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.AccountList, nil
}

// AccountFind finds an account either by name or fingerprint
func (c *Client) AccountFind(identifier string) (api.Account, error) {
	accounts, err := c.AccountList()
	if err != nil {
		return api.Account{}, err
	}

	for _, account := range accounts {
		if tools.GetCSRHash(account.CSR, true) == identifier {
			return account, nil
		}
		if account.Name == identifier {
			return account, nil
		}
	}

	return api.Account{}, ErrNotFound
}

// AccountAccept accepts a pending account
func (c *Client) AccountAccept(id uuid.UUID, expire time.Time) (api.Account, error) {
	var rep api.ControlAccountUpdateRep
	if err := c.conn.Request(
		api.ControlAccountUpdate,
		api.ControlAccountUpdateReq{
			Account: api.Account{
				ID:     id,
				Status: api.AccountStatusEnabled,
			},
			Expire: expire,
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(rep.Response)
	}
	return rep.Account, nil
}

// AccountReject rejects a pending account
func (c *Client) AccountReject(id uuid.UUID) (api.Account, error) {
	var rep api.ControlAccountUpdateRep
	if err := c.conn.Request(
		api.ControlAccountUpdate,
		api.ControlAccountUpdateReq{
			Account: api.Account{
				ID:     id,
				Status: api.AccountStatusDisabled,
			},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(rep.Response)
	}
	return rep.Account, nil
}

// AccountDelete deletes accounts
func (c *Client) AccountDelete(id uuid.UUID) error {
	var rep api.ControlAccountDeleteRep

	if err := c.conn.Request(
		api.ControlAccountDelete,
		api.ControlAccountDeleteReq(id),
		&rep, 5*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

// AccountRenewAPIKey deletes accounts
func (c *Client) AccountRenewAPIKey(id uuid.UUID) (api.Account, error) {
	var rep api.ControlAccountRenewAPIKeyRep

	if err := c.conn.Request(
		api.ControlAccountRenewAPIKey,
		api.ControlAccountRenewAPIKeyReq(id),
		&rep, 5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(rep.Response)
	}
	return rep.Account, nil
}

// PMProcessQuery returns a list of ended processes
func (c *Client) PMProcessQuery(level api.LogLevel, includeClosed bool, processIDs []uuid.UUID, returnLogs bool) ([]api.PMProcess, error) {
	var rep api.ControlPMProcessQueryRep
	if err := c.conn.Request(
		api.ControlPMProcessQuery,
		&api.ControlPMProcessQueryReq{
			Level:         level,
			IncludeClosed: includeClosed,
			ProcessIDs:    processIDs,
			ReturnLogs:    returnLogs,
		},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.PMProcessList, nil
}

// PMProcessSetStatus changes the postmortem status of a process
func (c *Client) PMProcessSetStatus(processID uuid.UUID, status api.PostMortemStatus, comment string) error {
	var rep api.Response
	if err := c.conn.Request(
		api.ControlPMProcessSetStatus,
		&api.ControlPMProcessSetStatusReq{
			ProcessID: processID,
			Status:    status,
			Comment:   comment,
		},
		&rep, 5*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep)
	}
	return nil
}

func (c *Client) request(subject string, req interface{}, rep interface{}, response *api.Response) error {
	if err := c.conn.Request(
		subject, req, rep, 5*time.Second,
	); err != nil {
		return err
	}
	if response.Status != api.StatusOK {
		return wrapAPIError(*response)
	}
	return nil
}

// LogsPurge purges logs
func (c *Client) LogsPurge(before time.Time) (int64, error) {
	var rep api.ControlLogsPurgeRep

	c.Log.Noticef("calling server api")
	err := c.request(
		api.ControlLogsPurge,
		&api.ControlLogsPurgeReq{
			Before: before,
		},
		&rep, &rep.Response,
	)

	return rep.Count, err
}
