// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"crypto"
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"time"

	"github.com/getsentry/raven-go"
	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// ServiceOptions holds options for an actor service
type ServiceOptions struct {
	Type     string
	Settings map[string]interface{}
}

// ActorOptions holds options for an actor
type ActorOptions struct {
	api.Actor `yaml:",inline"`
	Service   ServiceOptions
}

// PersistentStore is a key-value store that
type PersistentStore interface {
	Put(key string, value string) error
	Get(key string) (string, error)
	Del(key string) error
}

// COptions configures a Client
type COptions struct {
	ID   uuid.UUID
	Name string
	Type api.AccountType

	ForceStart bool

	PubKey  crypto.PublicKey
	PrivKey *ecdsa.PrivateKey

	Certificate []byte
	CSR         []byte
	RootCAs     [][]byte

	ActorList []ActorOptions

	ConnOptions nats.Options

	PersistentStore PersistentStore

	SentryDSN string

	// ChangeCallback is a function called whenever a change has occured on this set of options.
	// This happens, in particular, during account registration.
	// Optional.
	ChangeCallback func(COptions)
}

// NewClient creates a new client
func NewClient(options COptions, logger Logger) *Client {
	if options.ChangeCallback == nil {
		options.ChangeCallback = func(COptions) {}
	}
	if options.PersistentStore == nil {
		var err error
		options.PersistentStore, err = NewFilePersistentStore("pstore.db")
		if err != nil {
			panic(err)
		}
	}
	var ravenClient *raven.Client
	if options.SentryDSN != "" {
		var err error
		ravenClient, err = raven.New(options.SentryDSN)
		if err != nil {
			panic(err)
		}
		logger = NewRavenLogger(ravenClient, logger)
		logger.Noticef("Errors will be sent to sentry")
	}
	return &Client{
		options: options,
		Log:     logger,
		actors:  make(map[uuid.UUID]*Actor),
		stopped: make(chan bool),

		processingEndQueue:      make(chan processingEndEntry, 1),
		processingEndForceRetry: make(chan bool),

		ravenClient: ravenClient,
	}
}

// processingEndEntry
type processingEndEntry struct {
	actorID    uuid.UUID
	context    api.EnvelopeContext
	envelopeID uuid.UUID
	result     api.ActorProcessingResultType
	messages   []api.LogMessage
}

// Client is a xbus client api
type Client struct {
	options       COptions
	conn          *tools.Conn
	authenticated bool
	Log           Logger
	actors        map[uuid.UUID]*Actor
	running       []ActorService
	quit          chan bool
	stopped       chan bool

	processingEndQueue       chan processingEndEntry
	processingEndForceRetry  chan bool
	processingEndWaitingList []processingEndEntry

	ravenClient *raven.Client
}

// GetConn returns the underlying nats connection
func (c *Client) GetConn() *tools.Conn {
	return c.conn
}

// GetPersistentStore returns the persistentstore
func (c *Client) GetPersistentStore() PersistentStore {
	return c.options.PersistentStore
}

// ID returns the account ID (may be zero)
func (c *Client) ID() uuid.UUID {
	return c.options.ID
}

// Connect establises a connection to the bus
func (c *Client) Connect() error {
	return c.connect(nil)
}

func (c *Client) connect(disconnectedCB nats.ConnHandler) error {
	var ns string
	opts := c.options.ConnOptions
	opts.Secure = true
	opts.DisconnectedCB = disconnectedCB

	if c.options.Certificate == nil {
		c.Log.Noticef("Anonymous connection")
		opts.User = nuid.Next()
		ns = "anon-" + opts.User

	} else {
		c.Log.Debugf("Using TLS cert to connect")

		// Fill TLSConfig elements based on upper settings when necessary.
		if opts.TLSConfig == nil {
			opts.TLSConfig = &tls.Config{}
		}
		if opts.TLSConfig.Certificates == nil {
			opts.TLSConfig.Certificates = []tls.Certificate{tls.Certificate{
				Certificate: [][]byte{c.options.Certificate},
				PrivateKey:  c.options.PrivKey,
			}}
		}
		if opts.TLSConfig.RootCAs == nil {
			var rootCAs *x509.CertPool
			if c.options.RootCAs != nil {
				rootCAs = x509.NewCertPool()
				for _, ca := range c.options.RootCAs {
					cert, err := x509.ParseCertificate(ca)
					if err != nil {
						return fmt.Errorf("Cannot parse root ca: %s", err)
					}
					rootCAs.AddCert(cert)
				}
			}
			opts.TLSConfig.RootCAs = rootCAs
		}

		ns = c.options.ID.String()
	}

	conn, err := opts.Connect()
	if err != nil {
		return err
	}
	eConn, err := nats.NewEncodedConn(conn, "json")
	if err != nil {
		return err
	}

	c.conn = &tools.Conn{EncodedConn: eConn, Namespace: ns}

	c.authenticated = c.options.Certificate != nil

	c.Log.Noticef("Connected to %s", c.conn.Conn.ConnectedUrl())

	return nil
}

// Close closes the connection to the bus
func (c *Client) Close() {
	c.conn.Close()
	c.conn = nil
}

// AddActor attach an actor
func (c *Client) AddActor(actor ActorOptions) {
	c.options.ActorList = append(c.options.ActorList, actor)
}

// GetActorOptions returns an actor options from its id
func (c *Client) GetActorOptions(id uuid.UUID) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.ID == id {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// GetActorOptionsByName returns an actor options from its name
func (c *Client) GetActorOptionsByName(name string) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.Name == name {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// getActorFromOptions returns a *Actor from the options
func (c *Client) getActorFromOptions(options ActorOptions) *Actor {
	actor, ok := c.actors[options.ID]
	if !ok {
		actor = newActor(c, options)
		c.actors[options.ID] = actor
	}
	return actor
}

// GetActor returns a *Actor from its ID, or nil if it undefined
func (c *Client) GetActor(id uuid.UUID) *Actor {
	actor, ok := c.actors[id]
	if !ok {
		options, ok := c.GetActorOptions(id)
		if !ok {
			return nil
		}
		actor = c.getActorFromOptions(options)
	}
	return actor
}

// GetActorByName returns an actor from its name
func (c *Client) GetActorByName(name string) *Actor {
	actorOpts, ok := c.GetActorOptionsByName(name)
	if ok {
		actor, ok := c.actors[actorOpts.ID]
		if !ok {
			actor = c.getActorFromOptions(actorOpts)
		}
		return actor
	}
	return nil
}

// RawProcessingEnd calls ProcessingEnd server API directly
func (c *Client) RawProcessingEnd(
	actorID uuid.UUID,
	context api.EnvelopeContext,
	envelopeID uuid.UUID,
	result api.ActorProcessingResultType,
	messages []api.LogMessage,
) error {
	var rep api.ActorProcessingEndRep
	if err := c.GetConn().Request(
		api.ActorProcessingEnd(actorID),
		api.ActorProcessingEndReq{
			Context:    context,
			EnvelopeID: envelopeID,
			Result:     result,
			Messages:   messages,
		},
		&rep,
		10*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

func (c *Client) tryProcessingEnd(
	actorID uuid.UUID,
	context api.EnvelopeContext,
	envelopeID uuid.UUID,
	result api.ActorProcessingResultType,
	messages []api.LogMessage,
	firstAttempt bool,
) (tryAgain bool, err error) {
	key := actorID.String() +
		"/" + context.ProcessID.String() +
		"/" + context.NodeID +
		"/" + envelopeID.String() +
		"/" + "processingend"

	if firstAttempt {
		sResult, err := result.MarshalText()
		if err != nil {
			panic(err)
		}
		if err := c.GetPersistentStore().Put(key, string(sResult)); err != nil {
			c.Log.Errorf("Error using persistent store: %s.", err)
		}
	}

	if err := c.RawProcessingEnd(actorID, context, envelopeID, result, messages); err != nil {
		if err == nats.ErrTimeout ||
			err == nats.ErrConnectionClosed ||
			err == nats.ErrNoServers ||
			err == nats.ErrStaleConnection {
			// Most probably a temporary error (broker is down)
			// We will try again later in the background
			if firstAttempt {
				c.processingEndQueue <- processingEndEntry{
					actorID, context, envelopeID, result, messages,
				}
			}
			return true, err
		}
		return false, err
	}
	c.GetPersistentStore().Del(key)
	return false, nil
}

func (c *Client) processingEndWorker() {
	retry := make(chan bool)
	delay := time.Second

	doRetry := func() {
		if len(c.processingEndWaitingList) == 0 {
			return
		}
		var stillWaiting []processingEndEntry
		c.Log.Noticef("Attempt to send waiting ProcessingEnd(s)")
		for i, entry := range c.processingEndWaitingList {
			tryAgain, err := c.tryProcessingEnd(
				entry.actorID, entry.context, entry.envelopeID, entry.result,
				entry.messages, false)
			if err != nil {
				c.Log.Errorf("Error calling ProcessingEnd: %s", err)
			} else {
				delay = time.Second
			}
			if tryAgain {
				stillWaiting = c.processingEndWaitingList[i:]
				if delay < 5*time.Minute {
					delay = delay * 16 / 10
				}
				break
			}
		}
		if len(stillWaiting) != 0 {
			c.Log.Errorf("Will try again in '%f' seconds", delay.Seconds())
			time.AfterFunc(delay, func() { retry <- true })
		}
	}

	for {
		select {
		case newEntry := <-c.processingEndQueue:
			if len(c.processingEndWaitingList) == 0 {
				time.AfterFunc(delay, func() { retry <- true })
			}
			c.processingEndWaitingList = append(c.processingEndWaitingList, newEntry)
		case <-retry:
			doRetry()
		case <-c.processingEndForceRetry:
			delay = time.Second
			doRetry()
		case <-c.quit:
			return
		}
	}
}

// StartServices initialize and starts all the actors having a ActorService
func (c *Client) StartServices() error {
	var actors []ActorService
	for _, actorOpts := range c.options.ActorList {
		if actorOpts.Service.Type != "" {
			actors = append(actors, MakeActorService(
				actorOpts.Service.Type, c.getActorFromOptions(actorOpts)))
		}
	}
	for i, actor := range actors {
		if err := actor.Startup(); err != nil {
			c.Log.Errorf("Error starting actor %s: %s", c.options.ActorList[i].Name, err)
			if err := c.StopServices(); err != nil {
				c.Log.Errorf("Error shutting down actors: %s", err)
			}
			return err
		}
		c.running = append(c.running, actor)
	}
	return nil
}

// StopServices stops all the running actor
func (c *Client) StopServices() error {
	var lastError error
	actors := c.running[:]
	for _, actor := range actors {
		c.running = c.running[1:]
		if lastError = actor.Shutdown(); lastError != nil {
			c.Log.Errorf("Error shutting down actor: %s", lastError)
		}
	}
	return lastError
}

// Startup attempts to connect and start the services.
// If the connection attempt fail, it waits a little and
// retries forever.
// It blocks until a first successful connect + services start is performed*
func (c *Client) Startup() error {
	var (
		disconnected = make(chan bool)
		retry        = make(chan bool)
	)
	if c.ravenClient != nil {
		defer c.ravenClient.Wait()
	}
	err := c.connect(func(*nats.Conn) {
		disconnected <- true
	})
	if err != nil {
		if !c.options.ForceStart {
			return err
		}
		time.AfterFunc(time.Second, func() { retry <- true })
	} else {
		if err := c.StartServices(); err != nil {
			return err
		}
	}

	c.quit = make(chan bool)

	go c.processingEndWorker()

	go func() {
		defer func() {
			c.stopped <- true
		}()
		for {
			select {
			case <-disconnected:
				c.Log.Errorf("Got disconnected. Stopping services then reconnect")
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				c.Close()
			case <-c.quit:
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				if c.conn != nil {
					c.Close()
				}
				return
			case <-retry:
			}
			c.Log.Debugf("Reconnecting")
			if err := c.connect(func(*nats.Conn) {
				disconnected <- true
			}); err != nil {
				c.Log.Errorf("Error reconnecting: %s", err)
				time.AfterFunc(time.Second, func() { retry <- true })
				continue
			}
			c.Log.Debugf("Start services")
			err := c.StartServices()
			if err != nil {
				c.Log.Errorf("Error starting services: %s", err)
				c.Close()
				time.AfterFunc(time.Second, func() { retry <- true })
			}
			c.processingEndForceRetry <- true
		}
	}()
	return nil
}

// Shutdown stops the services if running and disconnect from
// the bus
func (c *Client) Shutdown() {
	close(c.quit)
	<-c.stopped
}

// ClientGetActors lists actors of the currently connected account
func (c *Client) ClientGetActors() ([]api.Actor, error) {
	var res api.ClientGetActorsRep
	if err := c.conn.Request(
		api.ClientSubject(c.ID(), api.ClientGetActors),
		api.ClientGetActorsReq(c.ID()),
		&res,
		5*time.Second,
	); err != nil {
		return nil, err
	}
	if res.Status != api.StatusOK {
		return nil, wrapAPIError(res.Response)
	}
	return res.Actors, nil
}

// WhoAmI returns informations on the currently connect account
// Works only when connected with a API Key
func (c *Client) WhoAmI(apiKey string) (api.Account, error) {
	var res api.WhoAmIRep
	whoamiSubject, whoamiReply := api.WhoAmISubjects(apiKey)
	if err := c.conn.RequestWithInbox(
		whoamiReply,
		whoamiSubject,
		"",
		&res,
		5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if res.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(res.Response)
	}
	c.options.ID = res.Account.ID
	c.options.Name = res.Account.Name
	c.options.Type = res.Account.Type

	c.conn.Namespace = c.options.ID.String()
	return res.Account, nil
}
